// Copyright (c) 2021, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

// https://github.com/vsergeev/c-periphery/blob/master/docs/gpio.md
// https://github.com/vsergeev/c-periphery/blob/master/src/gpio.c
// https://github.com/vsergeev/c-periphery/blob/master/src/gpio.h
// https://github.com/dart-lang/samples/tree/master/ffi

import 'dart:ffi';
import 'library.dart';
import 'package:ffi/ffi.dart';
import 'signature.dart';
import 'dart:convert';

/// Result codes of the [GPIO.poll].
enum GPIOpolling { SUCCESS, TIMEOUT }

/// Mapped native [GPIO] error codes with the same index, but different leading sign.
enum GPIOerrorCode {
  /// Error code for not able to map the native C enum
  ERROR_CODE_NOT_MAPPABLE,

  ///  Invalid arguments
  GPIO_ERROR_ARG,

  /// Opening GPIO
  GPIO_ERROR_OPEN,

  /// Line name not found
  GPIO_ERROR_NOT_FOUND,

  /// Querying GPIO attributes
  GPIO_ERROR_QUERY,

  /// Configuring GPIO attributes
  GPIO_ERROR_CONFIGURE,

  /// Unsupported attribute or operation
  GPIO_ERROR_UNSUPPORTED,

  /// Invalid operation
  GPIO_ERROR_INVALID_OPERATION,

  /// Reading/writing GPIO
  GPIO_ERROR_IO,

  /// Closing GPIO
  GPIO_ERROR_CLOSE,
}

/// [GPIO] input/output direction
enum GPIOdirection {
  ///  Input
  GPIO_DIR_IN,

  /// Output, initialized to low
  GPIO_DIR_OUT,

  /// output, initialized to low
  GPIO_DIR_OUT_LOW,

  /// output, initialized to high
  GPIO_DIR_OUT_HIGH
}

/// [GPIO] edge
enum GPIOedge {
  /// No interrupt edge
  GPIO_EDGE_NONE,

  /// Rising edge 0 -> 1
  GPIO_EDGE_RISING,

  /// Falling edge 1 -> 0
  GPIO_EDGE_FALLING,

  /// Both edges X -> !X
  GPIO_EDGE_BOTH
}

/// [GPIO] bias
enum GPIObias {
  /// Default line bias
  GPIO_BIAS_DEFAULT,

  /// Pull-up
  GPIO_BIAS_PULL_UP,

  /// Pull-down *
  GPIO_BIAS_PULL_DOWN,

  /// Disable line bias
  GPIO_BIAS_DISABLE,
}

/// [GPIO] drive
enum GPIOdrive {
  /// Default line drive (push-pull)
  GPIO_DRIVE_DEFAULT,

  /// Open drain
  GPIO_DRIVE_OPEN_DRAIN,

  ///  Open source
  GPIO_DRIVE_OPEN_SOURCE,
}

/// Result of the [GPIO.readEvent()].
class GPIOreadEvent {
  /// edge value
  final GPIOedge edge;
  // event time reported by Linux, in nanoseconds.
  final int nanoSeconds;

  GPIOreadEvent(this.edge, this.nanoSeconds);
}

/// Helper class for the static method [GPIO.pollMultiple].
///
/// [PollMultipleEvent.eventOccured] will be populated with true for the
/// corresponding GPIO in the gpios array if an edge event occurred, or false if none occurred.
/// This class contains also the number [eventCounter] of GPIOs for which an edge event occurred.
/// See for details [GPIO.pollMultiple]
class PollMultipleEvent {
  /// GPIO list monitored for edge events
  final List<GPIO> gpios;

  /// edge event occurred counter
  final int eventCounter;

  /// edge event result list
  final List<bool> eventOccured;
  PollMultipleEvent(this.gpios, this.eventCounter, this.eventOccured);

  /// Checks if the timeout has beeen reached.
  bool isTimeoutReached() {
    return eventCounter == 0 ? true : false;
  }

  /// Checks if an edge event occured for a [gpio]
  bool hasEventOccured(GPIO gpio) {
    var index = 0;
    for (var g in gpios) {
      if (g._gpioHandle == gpio._gpioHandle) {
        return eventOccured[index];
      }
      ++index;
    }
    return false;
  }
}

/* Configuration structure for gpio_open_*advanced() functions
typedef struct gpio_config {
    gpio_direction_t direction;
    gpio_edge_t edge;
    gpio_bias_t bias;
    gpio_drive_t drive;
    bool inverted;
    const char *label; 
} gpio_config_t;
*/

class _GPIOconfig extends Struct {
  @Int32()
  external int direction;
  @Int32()
  external int edge;
  @Int32()
  external int bias;
  @Int32()
  external int drive;
  @Int32() // this a bool value, but inside a struct 3 padding bytes are added for a 32-bit alignment
  external int inverted;
  external Pointer<Utf8> label;
}

/// Configuration class for [GPIO.advanced] and [GPIO.nameAdvanced].
class GPIOconfig {
  GPIOdirection direction;
  GPIOedge edge;
  GPIObias bias;
  GPIOdrive drive;
  bool inverted;
  // GPIO name
  String label;
  GPIOconfig(this.direction, this.edge, this.bias, this.drive, this.inverted,
      this.label);
  GPIOconfig.defaultValues()
      : direction = GPIOdirection.GPIO_DIR_IN,
        edge = GPIOedge.GPIO_EDGE_NONE,
        bias = GPIObias.GPIO_BIAS_DEFAULT,
        drive = GPIOdrive.GPIO_DRIVE_DEFAULT,
        inverted = false,
        label = '';
  Pointer<_GPIOconfig> toNative() {
    var cfg = malloc<_GPIOconfig>(1);
    cfg.ref.direction = direction.index;
    cfg.ref.edge = edge.index;
    cfg.ref.bias = bias.index;
    cfg.ref.drive = drive.index;
    cfg.ref.inverted = inverted ? 1 : 0;
    cfg.ref.label = label.toNativeUtf8();
    return cfg;
  }
}

/// [GPIO] exception
class GPIOexception implements Exception {
  final GPIOerrorCode errorCode;
  final String errorMsg;
  GPIOexception(this.errorCode, this.errorMsg);
  GPIOexception.errorCode(int code, Pointer<Void> handle)
      : errorCode = GPIO.getGPIOerrorCode(code),
        errorMsg = _getErrmsg(handle);
  @override
  String toString() => errorMsg;
}

final DynamicLibrary _peripheryLib = getPeripheryLib();

// gpio_t *gpio_new(void);
final _nativeGPIOnew = voidPtrVOIDM('gpio_new');

// int gpio_close(led_t *led);
final _nativeGPIOclose = intVoidM('gpio_close');

//  void gpio_free(gpio_t *i2c);
final _nativeGPIOfree = voidVoidM('gpio_free');

// int gpio_errno(gpio_t *i2c);
final _nativeGPIOerrno = intVoidM('gpio_errno');

// const char *gpio_errmsg(gpio_t *i2c);
final _nativeGPIOerrnMsg = utf8VoidM('gpio_errmsg');

// int gpio_tostring(gpio_t *gpio, char *str, size_t len);
final _nativeGPIOinfo = intVoidUtf8sizeTM('gpio_tostring');

// int gpio_fd(gpio_t *gpio);
final _nativeGPIOfd = intVoidM('gpio_fd');

// int gpio_chip_fd(gpio_t *gpio);
final _nativeGPIOchipFd = intVoidM('gpio_chip_fd');

// unsigned int gpio_line(gpio_t *gpio);
final _nativeGPIOline = intVoidM('gpio_line');

// int gpio_name(gpio_t *gpio, char *str, size_t len);
final _nativeGPIOname = intVoidUtf8sizeTM('gpio_name');

// int gpio_label(gpio_t *gpio, char *str, size_t len);
final _nativeGPIOlabel = intVoidUtf8sizeTM('gpio_label');

// int gpio_chip_name(gpio_t *gpio, char *str, size_t len);
final _nativeGPIOchipName = intVoidUtf8sizeTM('gpio_chip_name');

// int gpio_chip_label(gpio_t *gpio, char *str, size_t len);
final _nativeGPIOchipLabel = intVoidUtf8sizeTM('gpio_chip_label');

// int gpio_set_direction(gpio_t *gpio, gpio_direction_t direction);
final _nativeGPIOsetDirection = intVoidIntM('gpio_set_direction');

// int gpio_set_edge(gpio_t *gpio, gpio_edge_t edge);
final _nativeGPIOsetEdge = intVoidIntM('gpio_set_edge');

// int gpio_set_bias(gpio_t *gpio, gpio_bias_t bias);
final _nativeGPIOsetBias = intVoidIntM('gpio_set_bias');

// int gpio_set_drive(gpio_t *gpio, gpio_drive_t drive);
final _nativeGPIOsetDrive = intVoidIntM('gpio_set_drive');

// int gpio_set_inverted(gpio_t *gpio, bool inverted);
final _nativeGPIOsetInverted = intVoidBoolM('gpio_set_inverted');

// int gpio_get_direction(gpio_t *gpio, gpio_direction_t *direction);
final _nativeGPIOgetDirection = intVoidInt32PtrM('gpio_get_direction');

// int gpio_get_edge(gpio_t *gpio, gpio_edge_t *edge);
final _nativeGPIOgetEdge = intVoidInt32PtrM('gpio_get_edge');

// int gpio_get_bias(gpio_t *gpio, gpio_bias_t *bias);
final _nativeGPIOgetBias = intVoidInt32PtrM('gpio_get_bias');

// int gpio_get_drive(gpio_t *gpio, gpio_drive_t *drive);
final _nativeGPIOgetDrive = intVoidInt32PtrM('gpio_get_drive');

// int gpio_get_inverted(gpio_t *gpio, bool *inverted);
final _nativeGPIOgetInverted = intVoidInt8PtrM('gpio_get_inverted');

// int gpio_open(gpio_t *gpio, const char *path, unsigned int line, gpio_direction_t direction);
typedef _gpio_open = Int32 Function(
    Pointer<Void> handle, Pointer<Utf8> path, Int32 line, Int32 direction);
typedef _GPIOopen = int Function(
    Pointer<Void> handle, Pointer<Utf8> path, int line, int direction);
final _nativeGPIOopen = _peripheryLib
    .lookup<NativeFunction<_gpio_open>>('gpio_open')
    .asFunction<_GPIOopen>();

// int gpio_open_name(gpio_t *gpio, const char *path, const char *name, gpio_direction_t direction);
typedef _gpio_open_name = Int32 Function(Pointer<Void> handle,
    Pointer<Utf8> path, Pointer<Utf8> name, Int32 direction);
typedef _GPIOopenName = int Function(Pointer<Void> handle, Pointer<Utf8> path,
    Pointer<Utf8> name, int direction);
final _nativeGPIOopenName = _peripheryLib
    .lookup<NativeFunction<_gpio_open_name>>('gpio_open_name')
    .asFunction<_GPIOopenName>();

// int open_advanced(gpio_t *gpio, const char *path, unsigned int line, const gpio_config_t *config);
typedef _gpio_open_advanced = Int32 Function(Pointer<Void> handle,
    Pointer<Utf8> path, Int32 line, Pointer<_GPIOconfig> config);
typedef _GPIOopenAdvanced = int Function(Pointer<Void> handle,
    Pointer<Utf8> path, int line, Pointer<_GPIOconfig> config);
final _nativeGPIOopenAdvanced = _peripheryLib
    .lookup<NativeFunction<_gpio_open_advanced>>('gpio_open_advanced')
    .asFunction<_GPIOopenAdvanced>();

// int gpio_open_sysfs(gpio_t *gpio, unsigned int line, gpio_direction_t direction)
typedef _gpio_open_sysfs = Int32 Function(
    Pointer<Void> handle, Int32 line, Int32 direction);
typedef _GPIOopenSysfs = int Function(
    Pointer<Void> handle, int line, int direction);
final _nativeGPIOopenSysfs = _peripheryLib
    .lookup<NativeFunction<_gpio_open_sysfs>>('gpio_open_sysfs')
    .asFunction<_GPIOopenSysfs>();

// int gpio_open_name_advanced(gpio_t *gpio, const char *path, const char *name, const gpio_config_t *config);
typedef _gpio_open_name_advanced = Int32 Function(Pointer<Void> handle,
    Pointer<Utf8> path, Pointer<Utf8> name, Pointer<_GPIOconfig> config);
typedef _GPIOopenNameAdvanced = int Function(Pointer<Void> handle,
    Pointer<Utf8> path, Pointer<Utf8> name, Pointer<_GPIOconfig> config);
final _nativeGPIOopenNameAdvanced = _peripheryLib
    .lookup<NativeFunction<_gpio_open_name_advanced>>('gpio_open_name_advanced')
    .asFunction<_GPIOopenNameAdvanced>();

// int gpio_write(gpio_t *gpio, bool value)
final _nativeGPIOwrite = intVoidIntM('gpio_write');

// int gpio_read(gpio_t *gpio, bool *value);
final _nativeGPIOread = intVoidInt8PtrM('gpio_read');

// int gpio_read_event(gpio_t *gpio, gpio_edge_t *edge, uint64_t *timestamp);
typedef _gpio_read_event = Int32 Function(
    Pointer<Void>, Pointer<Int32> edge, Pointer<Uint64> timeoutMillis);
typedef _GPIOreadEvent = int Function(
    Pointer<Void>, Pointer<Int32> edge, Pointer<Uint64> timeoutMillis);
final _nativeGPIOReadEvent = _peripheryLib
    .lookup<NativeFunction<_gpio_read_event>>('gpio_read_event')
    .asFunction<_GPIOreadEvent>();

// int gpio_poll(gpio_t *gpio, int timeout_ms);
final _nativeGPIOpoll = intVoidIntM('gpio_poll');

// int gpio_poll_multiple(gpio_t **gpios, size_t count, int timeout_ms, bool *gpios_ready);
typedef _gpio_multiple_poll = Int32 Function(Pointer<Pointer<Void>> gpios,
    IntPtr count, Int32 timoutMillis, Pointer<Int8>);
typedef _GPIOmultiplePoll = int Function(
    Pointer<Pointer<Void>> gpios, int count, int timeoutMillis, Pointer<Int8>);
final _nativeGPIOmultiplePoll = _peripheryLib
    .lookup<NativeFunction<_gpio_multiple_poll>>('gpio_poll_multiple')
    .asFunction<_GPIOmultiplePoll>();

String _getErrmsg(Pointer<Void> handle) {
  return _nativeGPIOerrnMsg(handle).toDartString();
}

const BUFFER_LEN = 256;
const OPEN_ERROR = 'gpio_new() failed';

int _checkError(int value) {
  if (value < 0) {
    throw GPIOexception(
        GPIO.getGPIOerrorCode(value), GPIO.getGPIOerrorCode(value).toString());
  }
  return value;
}

final Map<String, dynamic> _map = {};

Map<String, dynamic> _jsonMap(String json) {
  if (_map.isEmpty) {
    _map.addAll(jsonDecode(json) as Map<String, dynamic>);
  }
  return _map;
}

/// GPIO wrapper functions for Linux userspace character device gpio-cdev and sysfs GPIOs.
///
/// Character device GPIOs were introduced in Linux kernel version 4.8. If the toolchain used to compiled
/// c-periphery contains Linux kernel headers older than 4.8 (i.e. linux/gpio.h is missing), then only legacy
/// sysfs GPIOs will be supported.
///
/// c-periphery [GPIO](https://github.com/vsergeev/c-periphery/blob/master/docs/gpio.md) documentation.
class GPIO {
  static String _gpioBasePath = '/dev/gpiochip';

  /// GPIO chip device path e.g. /dev/gpiochip0
  final String path;

  /// GPIO chip number, e.g. 0 for /dev/gpiochip0
  final int chip;

  /// GPIO line number, is -1 if  [GPIO.name] is used
  final int line;

  /// input/ouput GPIO direction
  final GPIOdirection direction;

  /// GPIO name, is empty if [GPIO.line] is used
  final String name;
  final Pointer<Void> _gpioHandle;
  bool _invalid = false;

  /// Conveverts a [GPIO] to a JSON string. See constructor [isolate] for detials.
  String toJson() {
    return '{"path":"$path","chip":$chip,"line":$line,"direction":${direction.index},"name":"$name","handle":${_gpioHandle.address}}';
  }

  /// Sets an alternative [chipBasePath], default value is '/dev/gpiochip'
  static void setBaseGPIOpath(String chipBasePath) {
    _gpioBasePath = chipBasePath;
  }

  /// Converts the native error code [value] to [GPIOerrorCode].
  static GPIOerrorCode getGPIOerrorCode(int value) {
    // must be negative
    if (value >= 0) {
      return GPIOerrorCode.ERROR_CODE_NOT_MAPPABLE;
    }
    value = -value;

    // check range
    if (value > GPIOerrorCode.GPIO_ERROR_CLOSE.index) {
      return GPIOerrorCode.ERROR_CODE_NOT_MAPPABLE;
    }

    return GPIOerrorCode.values[value];
  }

  void _checkStatus() {
    if (_invalid) {
      throw GPIOexception(GPIOerrorCode.GPIO_ERROR_INVALID_OPERATION,
          'GPIO line has the status released!');
    }
  }

  /// Returns the libc errno of the last failure that occurred.
  int getErrno() {
    _checkStatus();
    return _nativeGPIOerrno(_gpioHandle);
  }

  /// Opens the character device GPIO with the specified GPIO [line] and [direction] at
  /// the default character device GPIO with the [chip] number. The default
  /// chip number is 0, with the path /dev/gpiochip0.
  ///
  /// Use [GPIO.setBaseGPIOpath] to change the default character device path.
  GPIO(this.line, this.direction, [this.chip = 0])
      : path = _gpioBasePath + chip.toString(),
        name = '',
        _gpioHandle =
            _openGPIO(_gpioBasePath + chip.toString(), line, direction);

  static Pointer<Void> _openGPIO(
      String path, int line, GPIOdirection direction) {
    var _gpioHandle = _nativeGPIOnew();
    if (_gpioHandle == nullptr) {
      return throw GPIOexception(GPIOerrorCode.GPIO_ERROR_OPEN, OPEN_ERROR);
    }
    _checkError(_nativeGPIOopen(
        _gpioHandle, path.toNativeUtf8(), line, direction.index));
    return _gpioHandle;
  }

  /// Opens the character device GPIO with the specified GPIO [name] and [direction] at the default character
  /// device GPIO with the [chip] number. The default chip number is 0, with the path /dev/gpiochip0. Use [GPIO.setBaseGPIOpath]
  /// to change the default character device path.
  GPIO.name(this.name, this.direction, [this.chip = 0])
      : path = _gpioBasePath + chip.toString(),
        line = -1,
        _gpioHandle =
            _openNameGPIO(_gpioBasePath + chip.toString(), name, direction);

  static Pointer<Void> _openNameGPIO(
      String path, String name, GPIOdirection direction) {
    var _gpioHandle = _nativeGPIOnew();
    if (_gpioHandle == nullptr) {
      return throw GPIOexception(GPIOerrorCode.GPIO_ERROR_OPEN, OPEN_ERROR);
    }
    _checkError(_nativeGPIOopenName(_gpioHandle, path.toNativeUtf8(),
        name.toNativeUtf8(), direction.index));
    return _gpioHandle;
  }

  /// Opens the character device GPIO with the specified GPIO [line] and configuration [config] at the default character
  /// device GPIO with the [chip] number. The default chip numer is 0, with the path /dev/gpiochip0. Use [GPIO.setBaseGPIOpath]
  /// to change the default character device path.
  GPIO.advanced(this.line, GPIOconfig config, [this.chip = 0])
      : path = _gpioBasePath + chip.toString(),
        name = '',
        direction = config.direction,
        _gpioHandle =
            _openAdvancedGPIO(_gpioBasePath + chip.toString(), line, config);

  static Pointer<Void> _openAdvancedGPIO(
      String path, int line, GPIOconfig config) {
    var _gpioHandle = _nativeGPIOnew();
    if (_gpioHandle == nullptr) {
      return throw GPIOexception(GPIOerrorCode.GPIO_ERROR_OPEN, OPEN_ERROR);
    }
    _checkError(_nativeGPIOopenAdvanced(
        _gpioHandle, path.toNativeUtf8(), line, config.toNative()));
    return _gpioHandle;
  }

  /// Opens the character device GPIO with the specified GPIO [name] and the configuration [config] at the default character
  /// device GPIO with the [chip] number. The default chip numer is 0, with the path <tt>/dev/gpiochip0</tt>. Use [GPIO.setBaseGPIOpath]
  /// to change the default character device path.
  GPIO.nameAdvanced(this.name, GPIOconfig config, [this.chip = 0])
      : path = _gpioBasePath + chip.toString(),
        line = -1,
        direction = config.direction,
        _gpioHandle = _openNameAdvancedGPIO(
            _gpioBasePath + chip.toString(), name, config);

  static Pointer<Void> _openNameAdvancedGPIO(
      String path, String name, GPIOconfig config) {
    var _gpioHandle = _nativeGPIOnew();
    if (_gpioHandle == nullptr) {
      return throw GPIOexception(GPIOerrorCode.GPIO_ERROR_OPEN, OPEN_ERROR);
    }
    _checkError(_nativeGPIOopenNameAdvanced(_gpioHandle, path.toNativeUtf8(),
        name.toNativeUtf8(), config.toNative()));
    return _gpioHandle;
  }

  /// Opens the sysfs GPIO with the specified [line] and [direction].
  GPIO.sysfs(this.line, this.direction)
      : chip = -1,
        path = '',
        name = '',
        _gpioHandle = _openSysfsGPIO(line, direction);

  /// Duplicates an existing [GPIO] from a JSON string. This special constustor
  /// is used to transfer an existing [GPIO] to an other isolate.
  GPIO.isoloate(String json)
      : chip = _jsonMap(json)['chip'] as int,
        line = _jsonMap(json)['line'] as int,
        name = _jsonMap(json)['name'] as String,
        path = _jsonMap(json)['path'] as String,
        direction = GPIOdirection.values[_jsonMap(json)['direction'] as int],
        _gpioHandle =
            Pointer<Void>.fromAddress(_jsonMap(json)['handle'] as int);

  static Pointer<Void> _openSysfsGPIO(int line, GPIOdirection direction) {
    var _gpioHandle = _nativeGPIOnew();
    if (_gpioHandle == nullptr) {
      return throw GPIOexception(
          GPIOerrorCode.GPIO_ERROR_OPEN, 'Error opening GPIO interface');
    }
    _checkError(_nativeGPIOopenSysfs(_gpioHandle, line, direction.index));
    return _gpioHandle;
  }

  /// Polls multiple GPIOs for an edge event configured with [GPIO.setGPIOedge].
  /// For character device GPIOs, the edge event should be consumed with [GPIO.readEvent]. For sysfs GPIOs,
  /// the edge event should be consumed with [GPIO.read].
  /// [timeoutMillis] can be positive for a timeout in milliseconds, zero for a non-blocking poll, or
  /// negative for a blocking poll. Returns a [PollMultipleEvent()]
  static PollMultipleEvent pollMultiple(List<GPIO> gpios, int timeoutMillis) {
    final ptr = malloc<Pointer<Void>>(gpios.length);
    final result = malloc<Int8>(gpios.length);
    try {
      var index = 0;
      for (var g in gpios) {
        g._checkStatus();
        ptr[index++] = g._gpioHandle;
      }
      _checkError(
          _nativeGPIOmultiplePoll(ptr, gpios.length, timeoutMillis, result));

      var list = List<bool>.filled(gpios.length, false);
      var counter = 0;
      for (var i = 0; i < gpios.length; ++i) {
        list[i] = result[i] == 1 ? true : false;
        if (list[i]) {
          ++counter;
        }
      }
      return PollMultipleEvent(gpios, counter, list);
    } finally {
      malloc.free(ptr);
      malloc.free(result);
    }
  }

  /// Sets the state of the GPIO to [value].
  void write(bool value) {
    _checkStatus();
    _checkError(_nativeGPIOwrite(_gpioHandle, value ? 1 : 0));
  }

  /// Reads the state of the GPIO line.
  bool read() {
    return _getBoolValue(_nativeGPIOread);
  }

  /// Polls a GPIO for the edge event configured with [GPIO.setGPIOedge].
  /// For character device GPIOs, the edge event should be consumed with gpio_read_event().
  /// For sysfs GPIOs, the edge event should be consumed with [GPIO.read].
  GPIOpolling poll(int timeoutMillis) {
    _checkStatus();
    return _checkError(_nativeGPIOpoll(_gpioHandle, timeoutMillis)) == 1
        ? GPIOpolling.SUCCESS
        : GPIOpolling.TIMEOUT;
  }

  /// Reads the edge event that occurred with the GPIO.
  /// This method is intended for use with character device GPIOs and is unsupported by sysfs GPIOs.
  GPIOreadEvent readEvent() {
    _checkStatus();
    var edge = malloc<Int32>(1);
    var time = malloc<Uint64>(1);
    try {
      _checkError(_nativeGPIOReadEvent(_gpioHandle, edge, time));
      return GPIOreadEvent(GPIOedge.values[edge.value], time.value);
    } finally {
      malloc.free(edge);
      malloc.free(time);
    }
  }

  /// Returns the address of the internal handle.
  int getHandle() {
    return _gpioHandle.address;
  }

  /// Releases all internal native resoures.
  void dispose() {
    _checkStatus();
    _invalid = true;
    _checkError(_nativeGPIOclose(_gpioHandle));
    _nativeGPIOfree(_gpioHandle);
  }

  int _getInt32Value(intVoidInt32PtrF f) {
    _checkStatus();
    var data = malloc<Int32>(1);
    try {
      _checkError(f(_gpioHandle, data));
      return data[0];
    } finally {
      malloc.free(data);
    }
  }

  bool _getBoolValue(intVoidInt8PtrF f) {
    _checkStatus();
    var data = malloc<Int8>(1);
    try {
      _checkError(f(_gpioHandle, data));
      return data[0] != 0;
    } finally {
      malloc.free(data);
    }
  }

  /// Returns the property direction of the GPIO.
  GPIOdirection getGPIOdirection() {
    return GPIOdirection.values[_getInt32Value(_nativeGPIOgetDirection)];
  }

  /// Returns the property edge of the GPIO.
  GPIOedge getGPIOedge() {
    return GPIOedge.values[_getInt32Value(_nativeGPIOgetEdge)];
  }

  /// Returns the property bias of the GPIO.
  GPIObias getGPIObias() {
    return GPIObias.values[_getInt32Value(_nativeGPIOgetBias)];
  }

  /// Returns the property drive of the GPIO.
  GPIOdrive getGPIOdrive() {
    return GPIOdrive.values[_getInt32Value(_nativeGPIOgetDrive)];
  }

  /// Returns if the GPIO line is inverted,
  bool getGPIOinverted() {
    return _getBoolValue(_nativeGPIOgetInverted);
  }

  /// Sets the [direction] of the GPIO.
  void setGPIOdirection(GPIOdirection direction) {
    _checkStatus();
    _checkError(_nativeGPIOsetDirection(_gpioHandle, direction.index));
  }

  /// Sets the [edge] of the GPIO.
  void setGPIOedge(GPIOedge edge) {
    _checkStatus();
    _checkError(_nativeGPIOsetEdge(_gpioHandle, edge.index));
  }

  /// Sets the [bias] of the GPIO.
  void setGPIObias(GPIObias bias) {
    _checkStatus();
    _checkError(_nativeGPIOsetBias(_gpioHandle, bias.index));
  }

  /// Sets the [drive] of the GPIO.
  void setGPIOdrive(GPIOdrive drive) {
    _checkStatus();
    _checkError(_nativeGPIOsetDrive(_gpioHandle, drive.index));
  }

  /// Inverts the GPIO line.
  void setGPIOinverted(bool inverted) {
    _checkStatus();
    _checkError(_nativeGPIOsetInverted(_gpioHandle, inverted == true ? 1 : 0));
  }

  /// Returns the line the GPIO handle was opened with.
  int getLine() {
    _checkStatus();
    return _nativeGPIOline(_gpioHandle);
  }

  /// Returns the native line file descriptor of the GPIO handle.
  int getGPIOfd() {
    _checkStatus();
    return _nativeGPIOfd(_gpioHandle);
  }

  /// Returns the GPIO chip file descriptor of the GPIO handle.
  /// This method is intended for use with character device GPIOs and is unsupported by sysfs GPIOs.
  int getGPIOchipFD() {
    _checkStatus();
    return _nativeGPIOchipFd(_gpioHandle);
  }

  /// Returns the line name of the GPIO.
  /// This method is intended for use with character device GPIOs and always returns the empty string for sysfs GPIOs.
  String getGPIOname() {
    return _getString(_nativeGPIOname);
  }

  /// Returns the line consumer label of the GPIO.
  /// This method is intended for use with character device GPIOs and always returns the empty string for sysfs GPIOs.
  String getGPIOlabel() {
    return _getString(_nativeGPIOlabel);
  }

  /// Returns the label of the GPIO chip associated with the GPIO.
  String getGPIOchipName() {
    return _getString(_nativeGPIOchipName);
  }

  /// Returns the label of the GPIO chip associated with the GPIO.
  String getGPIOchipLabel() {
    return _getString(_nativeGPIOchipLabel);
  }

  String _getString(intVoidUtf8IntF f) {
    _checkStatus();
    var data = malloc<Int8>(BUFFER_LEN).cast<Utf8>();
    try {
      _checkError(f(_gpioHandle, data, BUFFER_LEN));
      return data.toDartString();
    } finally {
      malloc.free(data);
    }
  }

  /// Returns a string representation of the native GPIO handle.
  String getGPIOinfo() {
    return _getString(_nativeGPIOinfo);
  }
}
